/* 
 * tclCmdIL.c --
 *
 *	This file contains the top-level command routines for most of
 *	the Tcl built-in commands whose names begin with the letters
 *	I through L.  It contains only commands in the generic core
 *	(i.e. those that don't depend much upon UNIX facilities).
 *
 * Copyright (c) 1987-1993 The Regents of the University of California.
 * Copyright (c) 1994-1997 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclCmdIL.c 1.141 97/01/23 15:29:57
 */

#include "tclInt.h"
#include "tclPort.h"

/*
 * The following variable holds the full path name of the binary
 * from which this application was executed, or NULL if it isn't
 * know.  The value of the variable is set by the procedure
 * Tcl_FindExecutable.  The storage space is dynamically allocated.
 */

char *tclExecutableName = NULL;

/*
 * During execution of the "lsort" command, structures of the following
 * type are used to arrange the objects being sorted into a collection
 * of linked lists.
 */

typedef struct SortElement {
    Tcl_Obj *objPtr;			/* Object being sorted. */
    struct SortElement *nextPtr;        /* Next element in the list, or
					 * NULL for end of list. */
} SortElement;

/*
 * The "lsort" command needs to pass certain information down to the
 * function that compares two list elements, and the comparison function
 * needs to pass success or failure information back up to the top-level
 * "lsort" command.  The following structure is used to pass this
 * information.
 */

typedef struct SortInfo {
    int isIncreasing;		/* Nonzero means sort in increasing order. */
    int sortMode;		/* The sort mode.  One of SORTMODE_*
				 * values defined below */
    Tcl_DString compareCmd;	/* The Tcl comparison command when sortMode
				 * is SORTMODE_COMMAND.  Pre-initialized to
				 * hold base of command.*/
    long index;			/* If the -index option was specified, this
				 * holds the index of the list element
				 * to extract for comparison.  If -index
				 * wasn't specified, this is -1. */
    Tcl_Interp *interp;		/* The interpreter in which the sortis
				 * being done. */
    int resultCode;		/* Completion code for the lsort command.
				 * If an error occurs during the sort this
				 * is changed from TCL_OK to  TCL_ERROR. */
} SortInfo;

/*
 * The "sortMode" field of the SortInfo structure can take on any of the
 * following values.
 */

#define SORTMODE_ASCII      0
#define SORTMODE_INTEGER    1
#define SORTMODE_REAL       2
#define SORTMODE_COMMAND    3
#define SORTMODE_DICTIONARY 4

/*
 * Forward declarations for procedures defined in this file:
 */

static int		DictionaryCompare _ANSI_ARGS_((char *left,
			    char *right));
static SortElement *    MergeSort _ANSI_ARGS_((SortElement *headPt,
			    SortInfo *infoPtr));
static SortElement *    MergeLists _ANSI_ARGS_((SortElement *leftPtr,
			    SortElement *rightPtr, SortInfo *infoPtr));
static int		SortCompare _ANSI_ARGS_((Tcl_Obj *firstPtr,
			    Tcl_Obj *second, SortInfo *infoPtr));

/*
 *----------------------------------------------------------------------
 *
 * Tcl_IfCmd --
 *
 *	This procedure is invoked to process the "if" Tcl command.
 *	See the user documentation for details on what it does.
 *
 *	With the bytecode compiler, this procedure is only called when
 *	a command name is computed at runtime, and is "if" or the name
 *	to which "if" was renamed: e.g., "set z if; $z 1 {puts foo}"
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_IfCmd(dummy, interp, argc, argv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    int i, result, value;

    i = 1;
    while (1) {
	/*
	 * At this point in the loop, argv and argc refer to an expression
	 * to test, either for the main expression or an expression
	 * following an "elseif".  The arguments after the expression must
	 * be "then" (optional) and a script to execute if the expression is
	 * true.
	 */

	if (i >= argc) {
	    Tcl_AppendResult(interp, "wrong # args: no expression after \"",
		    argv[i-1], "\" argument", (char *) NULL);
	    return TCL_ERROR;
	}
	result = Tcl_ExprBoolean(interp, argv[i], &value);
	if (result != TCL_OK) {
	    return result;
	}
	i++;
	if ((i < argc) && (strcmp(argv[i], "then") == 0)) {
	    i++;
	}
	if (i >= argc) {
	    Tcl_AppendResult(interp, "wrong # args: no script following \"",
		    argv[i-1], "\" argument", (char *) NULL);
	    return TCL_ERROR;
	}
	if (value) {
	    return Tcl_Eval(interp, argv[i]);
	}
	
	/*
	 * The expression evaluated to false.  Skip the command, then
	 * see if there is an "else" or "elseif" clause.
	 */

	i++;
	if (i >= argc) {
	    return TCL_OK;
	}
	if ((argv[i][0] == 'e') && (strcmp(argv[i], "elseif") == 0)) {
	    i++;
	    continue;
	}
	break;
    }

    /*
     * Couldn't find a "then" or "elseif" clause to execute.  Check now
     * for an "else" clause.  We know that there's at least one more
     * argument when we get here.
     */

    if (strcmp(argv[i], "else") == 0) {
	i++;
	if (i >= argc) {
	    Tcl_AppendResult(interp,
		    "wrong # args: no script following \"else\" argument",
		    (char *) NULL);
	    return TCL_ERROR;
	}
    }
    return Tcl_Eval(interp, argv[i]);
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_IncrCmd --
 *
 *	This procedure is invoked to process the "incr" Tcl command.
 *	See the user documentation for details on what it does.
 *
 *	With the bytecode compiler, this procedure is only called when
 *	a command name is computed at runtime, and is "incr" or the name
 *	to which "incr" was renamed: e.g., "set z incr; $z i -1"
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
int
Tcl_IncrCmd(dummy, interp, argc, argv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    int value;
    char *oldString, *result;
    char newString[30];

    if ((argc != 2) && (argc != 3)) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" varName ?increment?\"", (char *) NULL);
	return TCL_ERROR;
    }

    oldString = Tcl_GetVar(interp, argv[1], TCL_LEAVE_ERR_MSG);
    if (oldString == NULL) {
	return TCL_ERROR;
    }
    if (Tcl_GetInt(interp, oldString, &value) != TCL_OK) {
	Tcl_AddErrorInfo(interp,
		"\n    (reading value of variable to increment)");
	return TCL_ERROR;
    }
    if (argc == 2) {
	value += 1;
    } else {
	int increment;

	if (Tcl_GetInt(interp, argv[2], &increment) != TCL_OK) {
	    Tcl_AddErrorInfo(interp,
		    "\n    (reading increment)");
	    return TCL_ERROR;
	}
	value += increment;
    }
    TclFormatInt(newString, value);
    result = Tcl_SetVar(interp, argv[1], newString, TCL_LEAVE_ERR_MSG);
    if (result == NULL) {
	return TCL_ERROR;
    }
    interp->result = result;
    return TCL_OK; 
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_InfoCmd --
 *
 *	This procedure is invoked to process the "info" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *---------------------------------------------------------------------- */

	/* ARGSUSED */
int
Tcl_InfoCmd(dummy, interp, argc, argv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
    register Interp *iPtr = (Interp *) interp;
    int length;
    Proc *procPtr;
    Var *varPtr;
    CompiledLocal *localPtr;
    Command *cmdPtr;
    Tcl_HashEntry *hPtr;
    Tcl_HashSearch search;
    int c, i;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" option ?arg arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }
    
    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'a') && (strncmp(argv[1], "args", (unsigned) length) == 0)) {
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " args procname\"", (char *) NULL);
	    return TCL_ERROR;
	}
	procPtr = TclFindProc(iPtr, argv[2]);
	if (procPtr == NULL) {
	    infoNoSuchProc:
	    Tcl_AppendResult(interp, "\"", argv[2],
		    "\" isn't a procedure", (char *) NULL);
	    return TCL_ERROR;
	}
	for (localPtr = procPtr->firstLocalPtr;  localPtr != NULL;
		localPtr = localPtr->nextPtr) {
	    if (localPtr->isArg) {
		Tcl_AppendElement(interp, localPtr->name);
	    }
	}
	return TCL_OK;
    } else if ((c == 'b')
	    && (strncmp(argv[1], "body", (unsigned) length) == 0)) {
	/*
	 * THIS FAILS IF THE OBJECT'S STRING REP HAS A NULL BYTE.
	 */
	
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		    " body procname\"", (char *) NULL);
	    return TCL_ERROR;
	}
	procPtr = TclFindProc(iPtr, argv[2]);
	if (procPtr == NULL) {
	    goto infoNoSuchProc;
	}
	iPtr->result = TclGetStringFromObj(procPtr->bodyPtr, &length);
	return TCL_OK;
    } else if ((c == 'c')
	    && (strncmp(argv[1], "cmdcount", (unsigned) length) == 0)
	    && (length >= 2)) {
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		    " cmdcount\"", (char *) NULL);
	    return TCL_ERROR;
	}
	TclFormatInt(iPtr->result, iPtr->cmdCount);
	return TCL_OK;
    } else if ((c == 'c')
	    && (strncmp(argv[1], "commands", (unsigned) length) == 0)
	    && (length >= 4)) {
	if (argc > 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		    " commands ?pattern?\"", (char *) NULL);
	    return TCL_ERROR;
	}
	for (hPtr = Tcl_FirstHashEntry(&iPtr->commandTable, &search);
		hPtr != NULL; hPtr = Tcl_NextHashEntry(&search)) {
	    char *name = Tcl_GetHashKey(&iPtr->commandTable, hPtr);
	    if ((argc == 3) && !Tcl_StringMatch(name, argv[2])) {
		continue;
	    }
	    Tcl_AppendElement(interp, name);
	}
	return TCL_OK;
    } else if ((c == 'c')
	    && (strncmp(argv[1], "complete", (unsigned) length) == 0)
	    && (length >= 4)) {
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		    " complete command\"", (char *) NULL);
	    return TCL_ERROR;
	}
	if (Tcl_CommandComplete(argv[2])) {
	    interp->result = "1";
	} else {
	    interp->result = "0";
	}
	return TCL_OK;
    } else if ((c == 'd')
	    && (strncmp(argv[1], "default", (unsigned) length) == 0)) {
	if (argc != 5) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " default procname arg varname\"",
		    (char *) NULL);
	    return TCL_ERROR;
	}
	procPtr = TclFindProc(iPtr, argv[2]);
	if (procPtr == NULL) {
	    goto infoNoSuchProc;
	}
	for (localPtr = procPtr->firstLocalPtr;  ;
	           localPtr = localPtr->nextPtr) {
	    if (localPtr == NULL) {
		Tcl_AppendResult(interp, "procedure \"", argv[2],
			"\" doesn't have an argument \"", argv[3],
			"\"", (char *) NULL);
		return TCL_ERROR;
	    }
	    if ((localPtr->isArg)
		    && (strcmp(argv[3], localPtr->name) == 0)) {
		Tcl_Obj *namePtr;
		int length = strlen(argv[4]);
		
		TclNewObj(namePtr);
		TclInitStringRep(namePtr, argv[4], (unsigned) length);
		if (localPtr->defValuePtr != NULL) {
		    if (Tcl_ObjSetVar2(interp, namePtr, NULL,
			    localPtr->defValuePtr, 0) == NULL) {
			defStoreError:
			Tcl_AppendResult(interp,
				"couldn't store default value in variable \"",
				argv[4], "\"", (char *) NULL);
			TclDecrRefCount(namePtr); /* done with name object */
			return TCL_ERROR;
		    }
		    iPtr->result = "1";
		} else {
		    Tcl_Obj *nullObjPtr, *resultObjPtr;
		    TclNewObj(nullObjPtr); /* created with ref count 1 */
		    resultObjPtr = Tcl_ObjSetVar2(interp, namePtr, NULL,
			    nullObjPtr, 0);
		    if (resultObjPtr == NULL) {
			TclDecrRefCount(nullObjPtr); /* done with object */
			goto defStoreError;
		    }
		    if (resultObjPtr == nullObjPtr) {
			/*
			 * Decrement the ref count for the null object back
			 * to 1 since it will have been incremented by
			 * Tcl_ObjSetVar2.
			 */
			
			TclDecrRefCount(nullObjPtr);
		    }
		    iPtr->result = "0";
		}
		TclDecrRefCount(namePtr); /* we're done with name object */
		return TCL_OK;
	    }
	}
    } else if ((c == 'e')
	    && (strncmp(argv[1], "exists", (unsigned) length) == 0)) {
	char *p;
	if (argc != 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		    " exists varName\"", (char *) NULL);
	    return TCL_ERROR;
	}
	p = Tcl_GetVar((Tcl_Interp *) iPtr, argv[2], 0);

	/*
	 * The code below handles the special case where the name is for
	 * an array: Tcl_GetVar will reject this since you can't read
	 * an array variable without an index.
	 */

	if (p == NULL) {
	    Tcl_HashEntry *hPtr;
	    Var *varPtr = NULL;

	    if (strchr(argv[2], '(') != NULL) {
		noVar:
		iPtr->result = "0";
		return TCL_OK;
	    }
	    if (iPtr->varFramePtr == NULL) {
		hPtr = Tcl_FindHashEntry(&iPtr->globalTable, argv[2]);
		if (hPtr != NULL) {
		    varPtr = (Var *) Tcl_GetHashValue(hPtr);
		}
	    } else {
		CallFrame *varFramePtr = iPtr->varFramePtr;
		int localVarCt = varFramePtr->procPtr->numCompiledLocals;
		Var *localVarPtr;
		
		for (i = 0, localVarPtr = varFramePtr->compiledLocals;
		        i < localVarCt;
			i++, localVarPtr++) {
		    if (strcmp(argv[2], localVarPtr->name) == 0) {
			varPtr = localVarPtr;
			break;
		    }
		}
		if ((varPtr == NULL)
		        && (varFramePtr->varTablePtr != NULL)) {
		    hPtr = Tcl_FindHashEntry(varFramePtr->varTablePtr,
					     argv[2]);
		    if (hPtr != NULL) {
			varPtr = (Var *) Tcl_GetHashValue(hPtr);
		    }
		}
	    }
	    if (varPtr == NULL) {
		goto noVar;
	    }
	    while (varPtr->flags & VAR_LINK) {
		varPtr = varPtr->value.linkPtr;
	    }
	    if (varPtr->flags & VAR_UNDEFINED) {
		goto noVar;
	    }
	    if (!(varPtr->flags & VAR_ARRAY)) {
		goto noVar;
	    }
	}
	iPtr->result = "1";
	return TCL_OK;
    } else if ((c == 'g')
	    && (strncmp(argv[1], "globals", (unsigned) length) == 0)) {
	char *name;

	if (argc > 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		    " globals ?pattern?\"", (char *) NULL);
	    return TCL_ERROR;
	}
	for (hPtr = Tcl_FirstHashEntry(&iPtr->globalTable, &search);
		hPtr != NULL; hPtr = Tcl_NextHashEntry(&search)) {
	    varPtr = (Var *) Tcl_GetHashValue(hPtr);
	    if (varPtr->flags & VAR_UNDEFINED) {
		continue;
	    }
	    name = Tcl_GetHashKey(&iPtr->globalTable, hPtr);
	    if ((argc == 3) && !Tcl_StringMatch(name, argv[2])) {
		continue;
	    }
	    Tcl_AppendElement(interp, name);
	}
	return TCL_OK;
    } else if ((c == 'h')
	    && (strncmp(argv[1], "hostname", (unsigned) length) == 0)) {
	if (argc > 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		    " hostname\"", (char *) NULL);
	    return TCL_ERROR;
	}
        Tcl_AppendResult(interp, Tcl_GetHostName(), NULL);
        return TCL_OK;
    } else if ((c == 'l')
	    && (strncmp(argv[1], "level", (unsigned) length) == 0)
	    && (length >= 2)) {
	if (argc == 2) {
	    if (iPtr->varFramePtr == NULL) {
		iPtr->result = "0";
	    } else {
		TclFormatInt(iPtr->result, iPtr->varFramePtr->level);
	    }
	    return TCL_OK;
	} else if (argc == 3) {
	    /*
	     * THIS FAILS IF AN ARG OBJECT'S STRING REP HAS A NULL BYTE.
	     */
	    
	    int level, i;
	    CallFrame *framePtr;
	    char **frameArgv;

	    if (Tcl_GetInt(interp, argv[2], &level) != TCL_OK) {
		return TCL_ERROR;
	    }
	    if (level <= 0) {
		if (iPtr->varFramePtr == NULL) {
		    levelError:
		    Tcl_AppendResult(interp, "bad level \"", argv[2],
			    "\"", (char *) NULL);
		    return TCL_ERROR;
		}
		level += iPtr->varFramePtr->level;
	    }
	    for (framePtr = iPtr->varFramePtr; framePtr != NULL;
		    framePtr = framePtr->callerVarPtr) {
		if (framePtr->level == level) {
		    break;
		}
	    }
	    if (framePtr == NULL) {
		goto levelError;
	    }
	    frameArgv = (char **)
		ckalloc((unsigned) framePtr->objc * sizeof(char *));
	    for (i = 0;  i < framePtr->objc;  i++) {
		Tcl_Obj *objPtr = framePtr->objv[i];
		frameArgv[i] = TclGetStringFromObj(objPtr, &length);
	    }
	    iPtr->result = Tcl_Merge(framePtr->objc, frameArgv);
	    iPtr->freeProc = TCL_DYNAMIC;
	    ckfree((char *) frameArgv);
	    return TCL_OK;
	}
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" level [number]\"", (char *) NULL);
	return TCL_ERROR;
    } else if ((c == 'l')
	    && (strncmp(argv[1], "library", (unsigned) length) == 0)
	    && (length >= 2)) {
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		    " library\"", (char *) NULL);
	    return TCL_ERROR;
	}
	interp->result = Tcl_GetVar(interp, "tcl_library", TCL_GLOBAL_ONLY);
	if (interp->result == NULL) {
	    interp->result = "no library has been specified for Tcl";
	    return TCL_ERROR;
	}
	return TCL_OK;
    } else if ((c == 'l')
	    && (strncmp(argv[1], "loaded", (unsigned) length) == 0)
	    && (length >= 3)) {
	if ((argc != 2) && (argc != 3)) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		    " loaded ?interp?\"", (char *) NULL);
	    return TCL_ERROR;
	}
	return TclGetLoadedPackages(interp, argv[2]);
    } else if ((c == 'l')
	    && (strncmp(argv[1], "locals", (unsigned) length) == 0)
	    && (length >= 3)) {
	int localVarCt;
	Var *localVarPtr;
	char *name;

	if (argc > 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		    " locals ?pattern?\"", (char *) NULL);
	    return TCL_ERROR;
	}
	if (iPtr->varFramePtr == NULL) {
	    return TCL_OK;
	}

	/*
	 * First add the compiled locals (i.e. the ones stored in the call
	 * frame), then add the ones in the hash table (if one exists).
	 */

	localVarCt = iPtr->varFramePtr->procPtr->numCompiledLocals;
	for (i = 0, localVarPtr = iPtr->varFramePtr->compiledLocals;
	            i < localVarCt;  i++, localVarPtr++) {
	    name = localVarPtr->name;
	    if ((argc == 3) && !Tcl_StringMatch(name, argv[2])) {
		continue;
	    }
	    Tcl_AppendElement(interp, name);
	}
	if (iPtr->varFramePtr->varTablePtr == NULL) {
	    return TCL_OK;
	}
	for (hPtr = Tcl_FirstHashEntry(iPtr->varFramePtr->varTablePtr, &search);
		hPtr != NULL;  hPtr = Tcl_NextHashEntry(&search)) {
	    varPtr = (Var *) Tcl_GetHashValue(hPtr);
	    if (varPtr->flags & (VAR_UNDEFINED|VAR_LINK)) {
		continue;
	    }
	    name = Tcl_GetHashKey(iPtr->varFramePtr->varTablePtr, hPtr);
	    if ((argc == 3) && !Tcl_StringMatch(name, argv[2])) {
		continue;
	    }
	    Tcl_AppendElement(interp, name);
	}
	return TCL_OK;
    } else if ((c == 'n')
	    && (strncmp(argv[1], "nameofexecutable", (unsigned) length) == 0)) {
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		    " nameofexecutable\"", (char *) NULL);
	    return TCL_ERROR;
	}
	if (tclExecutableName != NULL) {
	    interp->result = tclExecutableName;
	}
	return TCL_OK;
    } else if ((c == 'p')
	    && (strncmp(argv[1], "patchlevel", (unsigned) length) == 0)
	    && (length >= 2)) {
	char *value;

	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		    " patchlevel\"", (char *) NULL);
	    return TCL_ERROR;
	}
	value = Tcl_GetVar(interp, "tcl_patchLevel",
		TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG);
	if (value == NULL) {
	    return TCL_ERROR;
	}
	interp->result = value;
	return TCL_OK;
    } else if ((c == 'p')
	    && (strncmp(argv[1], "procs", (unsigned) length) == 0)
	    && (length >= 2)) {
	if (argc > 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		    " procs ?pattern?\"", (char *) NULL);
	    return TCL_ERROR;
	}
	for (hPtr = Tcl_FirstHashEntry(&iPtr->commandTable, &search);
		hPtr != NULL; hPtr = Tcl_NextHashEntry(&search)) {
	    char *name = Tcl_GetHashKey(&iPtr->commandTable, hPtr);

	    cmdPtr = (Command *) Tcl_GetHashValue(hPtr);
	    if (!TclIsProc(cmdPtr)) {
		continue;
	    }
	    if ((argc == 3) && !Tcl_StringMatch(name, argv[2])) {
		continue;
	    }
	    Tcl_AppendElement(interp, name);
	}
	return TCL_OK;
    } else if ((c == 's')
	    && (strncmp(argv[1], "script", (unsigned) length) == 0)
	    && (length >= 2)) {
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " script\"", (char *) NULL);
	    return TCL_ERROR;
	}
	if (iPtr->scriptFile != NULL) {
	    /*
	     * Can't depend on iPtr->scriptFile to be non-volatile:
	     * if this command is returned as the result of the script,
	     * then iPtr->scriptFile will go away.
	     */

	    Tcl_SetResult(interp, iPtr->scriptFile, TCL_VOLATILE);
	}
	return TCL_OK;
    } else if ((c == 's')
	    && (strncmp(argv[1], "sharedlibextension", (unsigned) length) == 0)
	    && (length >= 2)) {
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " sharedlibextension\"", (char *) NULL);
	    return TCL_ERROR;
	}
#ifdef TCL_SHLIB_EXT
	interp->result = TCL_SHLIB_EXT;
#endif
	return TCL_OK;
    } else if ((c == 't')
	    && (strncmp(argv[1], "tclversion", (unsigned) length) == 0)) {
	char *value;

	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " tclversion\"", (char *) NULL);
	    return TCL_ERROR;
	}
	value = Tcl_GetVar(interp, "tcl_version",
		TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG);
	if (value == NULL) {
	    return TCL_ERROR;
	}
	interp->result = value;
	return TCL_OK;
    } else if ((c == 'v')
	    && (strncmp(argv[1], "vars", (unsigned) length) == 0)) {
	Tcl_HashTable *tablePtr;
	Var *localVarPtr;
	char *name;

	if (argc > 3) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"",
		    argv[0], " vars ?pattern?\"", (char *) NULL);
	    return TCL_ERROR;
	}
	if (iPtr->varFramePtr != NULL) {
	    /*
	     * We're in a local frame. First add the compiled locals
	     * (i.e. the ones stored in the call frame), then add below the
	     * ones in the call frame's variable hash table (if one
	     * exists).
	     */

	    int localVarCt = iPtr->varFramePtr->procPtr->numCompiledLocals;
	    tablePtr = iPtr->varFramePtr->varTablePtr;
	    for (i = 0, localVarPtr = iPtr->varFramePtr->compiledLocals;
		     i < localVarCt;  i++, localVarPtr++) {
		if (localVarPtr->flags & VAR_UNDEFINED) {
		    continue;
		}
		name = localVarPtr->name;
		if ((argc == 3) && !Tcl_StringMatch(name, argv[2])) {
		    continue;
		}
		Tcl_AppendElement(interp, name);
	    }
	} else {
	    tablePtr = &iPtr->globalTable;
	}
	if (tablePtr == NULL) {
	    return TCL_OK;
	}
	for (hPtr = Tcl_FirstHashEntry(tablePtr, &search);
		hPtr != NULL;  hPtr = Tcl_NextHashEntry(&search)) {
	    varPtr = (Var *) Tcl_GetHashValue(hPtr);
	    if (varPtr->flags & VAR_UNDEFINED) {
		continue;
	    }
	    name = Tcl_GetHashKey(tablePtr, hPtr);
	    if ((argc == 3) && !Tcl_StringMatch(name, argv[2])) {
		continue;
	    }
	    Tcl_AppendElement(interp, name);
	}
	return TCL_OK;
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
		"\": should be args, body, cmdcount, commands, ",
		"complete, default, ",
		"exists, globals, hostname, level, library, loaded, locals, ",
		"nameofexecutable, patchlevel, procs, script, ",
		"sharedlibextension, tclversion, or vars",
		(char *) NULL);
	return TCL_ERROR;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_JoinObjCmd --
 *
 *	This procedure is invoked to process the "join" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_JoinObjCmd(dummy, interp, objc, objv)
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj **objv;		/* The argument objects. */
{
    Interp *iPtr = (Interp *) interp;
    Tcl_Obj *listPtr;
    List *listRepPtr;
    register Tcl_Obj **elemPtrs;
    register Tcl_Obj *elemPtr;
    register Tcl_DString *dsPtr;
    char *joinString, *bytes;
    int joinLength, listLen, length, i, result;
    Tcl_Obj *resultPtr = iPtr->objResult;

    if (objc == 2) {
	joinString = " ";
	joinLength = 1;
    } else if (objc == 3) {
	joinString = TclGetStringFromObj(objv[2], &joinLength);
    } else {
	Tcl_StringObjAppend(resultPtr, "wrong # args: should be \"", -1);
	Tcl_StringObjAppendObj(resultPtr, objv[0]);
	Tcl_StringObjAppend(resultPtr, " list ?joinString?\"", -1);
	return TCL_ERROR;
    }

    /*
     * Make sure the list argument is a list object and get its length.
     */

    listPtr = objv[1];
    if (listPtr->typePtr != &tclListType) {
	result = tclListType.setFromAnyProc(interp, listPtr);
	if (result != TCL_OK) {
	    return result;
	}
    }
    listRepPtr = (List *) listPtr->internalRep.otherValuePtr;
    listLen = listRepPtr->elemCount;

    /*
     * Now concatenate strings to form the "joined" result. We manipulate
     * the result string object's internal representation for speed. We
     * also append directly into the interpreter's result object.
     */

    dsPtr = (Tcl_DString *) ckalloc(sizeof(Tcl_DString));
    Tcl_DStringInit(dsPtr);

    elemPtrs = listRepPtr->elements;
    for (i = 0;  i < listLen;  i++) {
	if (i > 0) {
	    Tcl_DStringAppend(dsPtr, joinString, joinLength);
	}
	
	elemPtr = elemPtrs[i];
	bytes = TclGetStringFromObj(elemPtr, &length);
	Tcl_DStringAppend(dsPtr, bytes, length); /* adds NULL after end */
    }
    
    resultPtr->internalRep.otherValuePtr = (VOID*) dsPtr;
    resultPtr->typePtr = &tclStringType;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_LindexObjCmd --
 *
 *	This object-based procedure is invoked to process the "lindex" Tcl
 *	command. See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

    /* ARGSUSED */
int
Tcl_LindexObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    Tcl_Obj **objv;			/* Argument objects. */
{
    Interp *iPtr = (Interp *) interp;
    Tcl_Obj *listPtr;
    register List *listRepPtr;
    register Tcl_Obj *indexPtr;
    register char *indexStr;
    int index, length, result;
    Tcl_Obj *elemPtr, *oldResultPtr;

    if (objc != 3) {
        Tcl_StringObjAppend(iPtr->objResult, "wrong # args: should be \"", -1);
        Tcl_StringObjAppendObj(iPtr->objResult, objv[0]);
        Tcl_StringObjAppend(iPtr->objResult, " list index\"", -1);
	return TCL_ERROR;
    }

    listPtr = objv[1];
    if (listPtr->typePtr != &tclListType) {
	result = tclListType.setFromAnyProc(interp, listPtr);
	if (result != TCL_OK) {
	    return result;
	}
    }
    listRepPtr = (List *) listPtr->internalRep.otherValuePtr;

    /*
     * THIS FAILS IF THE INDEX OBJECT'S STRING REP CONTAINS NULLS.
     */

    indexPtr = objv[2];
    indexStr = TclGetStringFromObj(indexPtr, &length);
    if ((*indexStr == 'e')
	    && (strncmp(indexStr, "end", (unsigned) length) == 0)) {
	index = (listRepPtr->elemCount - 1);
    } else {
	if (indexPtr->typePtr != &tclIntType) {
	    result = tclIntType.setFromAnyProc(interp, indexPtr);
	    if (result != TCL_OK) {
		return result;
	    }
	}
	index = indexPtr->internalRep.intValue;
    }

    /*
     * Now fetch the index-th list element.
     */
    
    if ((index < 0) || (index >= listRepPtr->elemCount)) {
	/*
	 * The index is out of range: the result is an empty string object.
	 */
	
	return TCL_OK;
    }
    elemPtr = listRepPtr->elements[index];

    /*
     * Set the interpreter's object result. 
     */

    oldResultPtr = iPtr->objResult;
    if (oldResultPtr != NULL) {
	TclDecrRefCount(oldResultPtr);
    }
    iPtr->objResult = elemPtr;
    Tcl_IncrRefCount(elemPtr); /* iPtr->objResult is another ref */
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_LinsertObjCmd --
 *
 *	This object-based procedure is invoked to process the "linsert" Tcl
 *	command. See the user documentation for details on what it does.
 *
 * Results:
 *	A new Tcl list object formed by inserting zero or more elements 
 *	into a list.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_LinsertObjCmd(dummy, interp, objc, objv)
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    register int objc;		/* Number of arguments. */
    Tcl_Obj **objv;		/* Argument objects. */
{
    Interp *iPtr = (Interp *) interp;
    Tcl_Obj *listPtr;
    register Tcl_Obj *indexPtr;
    register char *indexStr;
    int index, length, isDuplicate;
    int result = TCL_OK;	/* Initialized to avoid compiler warning. */
    Tcl_Obj *resultPtr = iPtr->objResult;

    if (objc < 4) {
        Tcl_StringObjAppend(resultPtr, "wrong # args: should be \"", -1);
        Tcl_StringObjAppendObj(resultPtr, objv[0]);
        Tcl_StringObjAppend(resultPtr,
                 " list index element ?element ...?\"", -1);
	return TCL_ERROR;
    }

    /*
     * If the list object is unshared we can modify it directly. Otherwise
     * we create a copy to modify: this is "copy on write". We create the
     * copy directly in the interpreter's object result.
     */
    
    listPtr = objv[1];
    if (Tcl_IsShared(listPtr)) {
	Tcl_ObjType *typePtr = listPtr->typePtr;
	
	if (typePtr != NULL) {
	    typePtr->dupIntRepProc(listPtr, resultPtr);
	} else if (listPtr->bytes != NULL) {
	    TclInitStringRep(resultPtr, listPtr->bytes, listPtr->length);
	}
	listPtr = resultPtr;
	isDuplicate = 1;
    } else {
	isDuplicate = 0;
    }
    
    /*
     * THIS FAILS IF THE OBJECT'S STRING REP CONTAINS NULLS.
     */
    
    indexStr = TclGetStringFromObj(objv[2], &length);
    if ((*indexStr == 'e') && (strcmp(indexStr, "end") == 0)) {
	index = INT_MAX;
    } else {
	indexPtr = objv[2];
	if (indexPtr->typePtr != &tclIntType) {
	    result = tclIntType.setFromAnyProc(interp, indexPtr);
	    if (result != TCL_OK) {
		return result;
	    }
	}
	index = indexPtr->internalRep.intValue;
    }

    if ((objc == 4) && (index == INT_MAX)) {
	/*
	 * A special case: inserting one element at the end of the list.
	 */

	result = Tcl_ListObjAppendElement(interp, listPtr, objv[3]);
    } else if (objc > 3) {
	result = Tcl_ListObjReplace(interp, listPtr, index, 0,
				    (objc-3), &(objv[3]));
    }
    if (result != TCL_OK) {
	return result;
    }
    
    /*
     * Set the interpreter's object result.
     */

    if (!isDuplicate) {
	TclDecrRefCount(resultPtr);
	iPtr->objResult = listPtr;
	Tcl_IncrRefCount(listPtr); /* since interpreter now refers to it */
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_ListObjCmd --
 *
 *	This procedure is invoked to process the "list" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_ListObjCmd(dummy, interp, objc, objv)
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj **objv;		/* The argument objects. */
{
    Interp *iPtr = (Interp *) interp;
    Tcl_Obj *resultPtr = iPtr->objResult;
    register Tcl_Obj **elemPtrs;
    register Tcl_Obj *objPtr;
    List *listRepPtr;
    register int i;
    int numElems;

    /*
     * Handle first the special cases of 1) no arguments [the result is
     * an empty string object] and 2) a list of single object. In the
     * second case, we append directly into the interpreter's result object.
     */
    
    if (objc == 1) {
	return TCL_OK;
    } else if (objc == 2) {
	return Tcl_ListObjAppendElement(interp, iPtr->objResult, objv[1]);
    }
    
    /*
     * Otherwise set the interpreter's object result to a new list holding
     * the argument objects. We modify the interpreter's result object
     * directly.
     */

    numElems = (objc - 1);
    elemPtrs = (Tcl_Obj **)
	    ckalloc((unsigned) (numElems * sizeof(Tcl_Obj *)));
    for (i = 0;  i < numElems;  i++) {
	objPtr = objv[i+1];
	Tcl_IncrRefCount(objPtr);
	elemPtrs[i] = objPtr;
    }
    
    listRepPtr = (List *) ckalloc(sizeof(List));
    listRepPtr->maxElemCount = numElems;
    listRepPtr->elemCount    = numElems;
    listRepPtr->elements     = elemPtrs;
    
    resultPtr->internalRep.otherValuePtr = (VOID *) listRepPtr;
    resultPtr->typePtr = &tclListType;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_LlengthObjCmd --
 *
 *	This object-based procedure is invoked to process the "llength" Tcl
 *	command.  See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_LlengthObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    Tcl_Obj **objv;			/* Argument objects. */
{
    Interp *iPtr = (Interp *) interp;
    register Tcl_Obj *listPtr, *resultPtr;
    register List *listRepPtr;
    int result;

    if (objc != 2) {
        Tcl_StringObjAppend(iPtr->objResult, "wrong # args: should be \"", -1);
        Tcl_StringObjAppendObj(iPtr->objResult, objv[0]);
        Tcl_StringObjAppend(iPtr->objResult, " list\"", -1);
	return TCL_ERROR;
    }

    listPtr = objv[1];
    if (listPtr->typePtr != &tclListType) {
	result = tclListType.setFromAnyProc(interp, listPtr);
	if (result != TCL_OK) {
	    return result;
	}
    }
    listRepPtr = (List *) listPtr->internalRep.otherValuePtr;

    /*
     * Set the interpreter's object result to an integer object holding the
     * length. Note that we don't bother generating a string rep.
     * We modify the interpreter's result object directly.
     */

    resultPtr = iPtr->objResult;
    resultPtr->internalRep.intValue = listRepPtr->elemCount;
    resultPtr->typePtr = &tclIntType;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_LrangeObjCmd --
 *
 *	This procedure is invoked to process the "lrange" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_LrangeObjCmd(notUsed, interp, objc, objv)
    ClientData notUsed;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj **objv;		/* Argument objects. */
{
    Interp *iPtr = (Interp *) interp;
    Tcl_Obj *listPtr, *indexPtr;
    List *listRepPtr, *resultListRepPtr;
    register Tcl_Obj **elemPtrs, **resultElemPtrs;
    register Tcl_Obj *objPtr; 
    char *indexStr;
    int first, last, listLen, numElems, length, i, j, result;

    if (objc != 4) {
        Tcl_StringObjAppend(iPtr->objResult, "wrong # args: should be \"", -1);
        Tcl_StringObjAppendObj(iPtr->objResult, objv[0]);
        Tcl_StringObjAppend(iPtr->objResult, " list first last\"", -1);
	return TCL_ERROR;
    }

    /*
     * Make sure the list argument is a list object and get its length.
     */

    listPtr = objv[1];
    if (listPtr->typePtr != &tclListType) {
	result = tclListType.setFromAnyProc(interp, listPtr);
	if (result != TCL_OK) {
	    return result;
	}
    }
    listRepPtr = (List *) listPtr->internalRep.otherValuePtr;
    listLen = listRepPtr->elemCount;

    /*
     * THIS FAILS IF AN ARGUMENT INDEX OBJECT'S STRING REP CONTAINS NULLS.
     */

    indexPtr = objv[2];
    indexStr = TclGetStringFromObj(indexPtr, &length);
    if ((*indexStr == 'e')
	    && (strncmp(indexStr, "end", (unsigned) length) == 0)) {
	first = (listLen - 1);	/* index of the last element */
    } else {
	if (indexPtr->typePtr != &tclIntType) {
	    result = tclIntType.setFromAnyProc(interp, indexPtr);
	    if (result != TCL_OK) {
	        badIndex:
		Tcl_ResetObjResult(interp);
		Tcl_StringObjAppend(iPtr->objResult,
		        "expected integer or \"end\" but got \"", -1);
		Tcl_StringObjAppendObj(iPtr->objResult, indexPtr);
		Tcl_StringObjAppend(iPtr->objResult, "\"", -1);
		return result;
	    }
	}
	first = indexPtr->internalRep.intValue;
    }
    if (first < 0) {
	first = 0;
    }

    indexPtr = objv[3];
    indexStr = TclGetStringFromObj(indexPtr, &length);
    if ((*indexStr == 'e')
	    && (strncmp(indexStr, "end", (unsigned) length) == 0)) {
	last = (listLen - 1);
    } else {
	if (indexPtr->typePtr != &tclIntType) {
	    result = tclIntType.setFromAnyProc(interp, indexPtr);
	    if (result != TCL_OK) {
		goto badIndex;
	    }
	}
	last = indexPtr->internalRep.intValue;
	if (last >= listLen) {
	    last = (listLen - 1);
	}
    }
    if (first > last) {
	return TCL_OK;		/* the result is an empty object */
    }

    /*
     * Extract a range of fields. We directly access the list object's
     * internal representation for speed. We also append directly into
     * the interpreter's result object.
     */

    elemPtrs = listRepPtr->elements;
    numElems = (last - first + 1);
    resultElemPtrs = (Tcl_Obj **)
	    ckalloc((unsigned) (numElems * sizeof(Tcl_Obj *)));
    for (i = first, j = 0;  i <= last;  i++, j++) {
	objPtr = elemPtrs[i];
	Tcl_IncrRefCount(objPtr); /* since obj result now points to it */
	resultElemPtrs[j] = objPtr;
    }
    
    resultListRepPtr = (List *) ckalloc(sizeof(List));
    resultListRepPtr->maxElemCount = numElems;
    resultListRepPtr->elemCount    = numElems;
    resultListRepPtr->elements     = resultElemPtrs;
    
    iPtr->objResult->internalRep.otherValuePtr = (VOID *) resultListRepPtr;
    iPtr->objResult->typePtr = &tclListType;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_LreplaceObjCmd --
 *
 *	This object-based procedure is invoked to process the "lreplace" 
 *	Tcl command. See the user documentation for details on what it does.
 *
 * Results:
 *	A new Tcl list object formed by replacing zero or more elements of
 *	a list.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_LreplaceObjCmd(dummy, interp, objc, objv)
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj **objv;		/* Argument objects. */
{
    Interp *iPtr = (Interp *) interp;
    Tcl_Obj *listPtr;
    List *listRepPtr;
    register Tcl_Obj *firstPtr, *lastPtr;
    char *firstStr, *lastStr;
    int first, last, listLen, numToDelete, length, isDuplicate, result;
    Tcl_Obj *resultPtr = iPtr->objResult;

    if (objc < 4) {
        Tcl_StringObjAppend(resultPtr, "wrong # args: should be \"", -1);
        Tcl_StringObjAppendObj(resultPtr, objv[0]);
        Tcl_StringObjAppend(resultPtr,
                 " list first last ?element element ...?\"", -1);
	return TCL_ERROR;
    }

    /*
     * If the list object is unshared we can modify it directly, otherwise
     * we create a copy to modify: this is "copy on write". We create the
     * copy directly in the interpreter's object result.
     */
    
    listPtr = objv[1];
    if (Tcl_IsShared(listPtr)) {
	Tcl_ObjType *typePtr = listPtr->typePtr;
	
	if (typePtr != NULL) {
	    typePtr->dupIntRepProc(listPtr, resultPtr);
	} else if (listPtr->bytes != NULL) {
	    TclInitStringRep(resultPtr, listPtr->bytes, listPtr->length);
	}
	listPtr = resultPtr;
	isDuplicate = 1;
    } else {
	isDuplicate = 0;
    }

    if (listPtr->typePtr != &tclListType) {
	int result = tclListType.setFromAnyProc(interp, listPtr);
	if (result != TCL_OK) {
	    return result;
	}
    }
    listRepPtr = (List *) listPtr->internalRep.otherValuePtr;
    listLen = listRepPtr->elemCount;

    /*
     * THIS FAILS IF THE OBJECT'S STRING REP CONTAINS NULLS.
     */
    
    firstStr = TclGetStringFromObj(objv[2], &length);
    if ((*firstStr == 'e') && (strcmp(firstStr, "end") == 0)) {
	first = (listLen - 1);
    } else {
	firstPtr = objv[2];
	if (firstPtr->typePtr != &tclIntType) {
	    result = tclIntType.setFromAnyProc(interp, firstPtr);
	    if (result != TCL_OK) {
		Tcl_ResetObjResult(interp);
		Tcl_StringObjAppend(resultPtr, "bad index \"", -1);
		Tcl_StringObjAppend(resultPtr, firstStr, length);
		Tcl_StringObjAppend(resultPtr,
                        "\": must be integer or \"end\"", -1);
		return result;
	    }
	}
	first = firstPtr->internalRep.intValue;
    }
    
    /*
     * THIS FAILS IF THE OBJECT'S STRING REP CONTAINS NULLS.
     */
    
    lastStr = TclGetStringFromObj(objv[3], &length);
    if ((*lastStr == 'e') && (strcmp(lastStr, "end") == 0)) {
	last = (listLen - 1);
    } else {
	lastPtr = objv[3];
	if (lastPtr->typePtr != &tclIntType) {
	    result = tclIntType.setFromAnyProc(interp, lastPtr);
	    if (result != TCL_OK) {
		Tcl_ResetObjResult(interp);
		Tcl_StringObjAppend(resultPtr, "bad index \"", -1);
		Tcl_StringObjAppend(resultPtr, lastStr, length);
		Tcl_StringObjAppend(resultPtr,
                        "\": must be integer or \"end\"", -1);
		return result;
	    }
	}
	last = lastPtr->internalRep.intValue;
    }

    if (first < 0)  {
    	first = 0;
    }
    if (first >= listLen) {
	Tcl_ResetObjResult(interp);
	resultPtr = iPtr->objResult;
	Tcl_StringObjAppend(resultPtr, "list doesn't contain element ", -1);
	Tcl_StringObjAppendObj(resultPtr, objv[2]);
	return TCL_ERROR;
    }

    if (last >= listLen) {
    	last = (listLen - 1);
    }
    if (first <= last) {
	numToDelete = (last - first + 1);
    } else {
	numToDelete = 0;
    }

    if (objc > 4) {
	result = Tcl_ListObjReplace(interp, listPtr, first, numToDelete,
	        (objc-4), &(objv[4]));
    } else {
	result = Tcl_ListObjReplace(interp, listPtr, first, numToDelete,
		0, NULL);
    }
    if (result != TCL_OK) {
	return result;
    }

    /*
     * Set the interpreter's object result. 
     */

    if (!isDuplicate) {
	TclDecrRefCount(resultPtr);
	iPtr->objResult = listPtr;
	Tcl_IncrRefCount(listPtr); /* since interpreter now refers to it */
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_LsearchCmd --
 *
 *	This procedure is invoked to process the "lsearch" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
int
Tcl_LsearchCmd(notUsed, interp, argc, argv)
    ClientData notUsed;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
#define EXACT	0
#define GLOB	1
#define REGEXP	2
    int listArgc;
    char **listArgv;
    int i, match, mode, index;

    mode = GLOB;
    if (argc == 4) {
	if (strcmp(argv[1], "-exact") == 0) {
	    mode = EXACT;
	} else if (strcmp(argv[1], "-glob") == 0) {
	    mode = GLOB;
	} else if (strcmp(argv[1], "-regexp") == 0) {
	    mode = REGEXP;
	} else {
	    Tcl_AppendResult(interp, "bad search mode \"", argv[1],
		    "\": must be -exact, -glob, or -regexp", (char *) NULL);
	    return TCL_ERROR;
	}
    } else if (argc != 3) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
		" ?mode? list pattern\"", (char *) NULL);
	return TCL_ERROR;
    }
    if (Tcl_SplitList(interp, argv[argc-2], &listArgc, &listArgv) != TCL_OK) {
	return TCL_ERROR;
    }
    index = -1;
    for (i = 0; i < listArgc; i++) {
	match = 0;
	switch (mode) {
	    case EXACT:
		match = (strcmp(listArgv[i], argv[argc-1]) == 0);
		break;
	    case GLOB:
		match = Tcl_StringMatch(listArgv[i], argv[argc-1]);
		break;
	    case REGEXP:
		match = Tcl_RegExpMatch(interp, listArgv[i], argv[argc-1]);
		if (match < 0) {
		    ckfree((char *) listArgv);
		    return TCL_ERROR;
		}
		break;
	}
	if (match) {
	    index = i;
	    break;
	}
    }
    TclFormatInt(interp->result, index);
    ckfree((char *) listArgv);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_LsortObjCmd --
 *
 *	This procedure is invoked to process the "lsort" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

int
Tcl_LsortObjCmd(clientData, interp, objc, objv)
    ClientData clientData;		/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    Tcl_Obj **objv;			/* Argument values. */
{
    int i, index, dummy;
    Tcl_Obj *resultPtr;
    int length;
    Tcl_Obj *cmdPtr, **listObjPtrs;
    SortElement *elementArray;
    SortElement *elementPtr;        
    SortInfo sortInfo;                  /* Information about this sort that
                                         * needs to be passed to the 
                                         * comparison function */
    static char *switches[] =
	    {"-ascii", "-command", "-decreasing", "-dictionary",
	    "-increasing", "-index", "-integer", "-real", (char *) NULL};

    resultPtr = Tcl_GetObjResult(interp);
    if (objc < 2) {
	Tcl_StringObjAppend(resultPtr, "wrong # args: should be \"", -1);
	Tcl_StringObjAppendObj(resultPtr, objv[0]);
	Tcl_StringObjAppend(resultPtr, " ?options? list\"", -1);
	return TCL_ERROR;
    }

    /*
     * Parse arguments to set up the mode for the sort.
     */

    sortInfo.isIncreasing = 1;
    sortInfo.sortMode = SORTMODE_ASCII;
    sortInfo.index = -1;
    sortInfo.interp = interp;
    sortInfo.resultCode = TCL_OK;
    cmdPtr = NULL;
    for (i = 1; i < objc-1; i++) {
	if (Tcl_GetIndexFromObj(interp, objv[i], switches, "option", &index)
		!= TCL_OK) {
	    return TCL_ERROR;
	}
	switch (index) {
	    case 0:			/* -ascii */
		sortInfo.sortMode = SORTMODE_ASCII;
		break;
	    case 1:			/* -command */
		if (i == (objc-2)) {
		    Tcl_StringObjAppend(resultPtr,
			    "\"-command\" option must be followed by comparison command",
			    -1);
		    return TCL_ERROR;
		}
		sortInfo.sortMode = SORTMODE_COMMAND;
		cmdPtr = objv[i+1];
		i++;
		break;
	    case 2:			/* -decreasing */
		sortInfo.isIncreasing = 0;
		break;
	    case 3:			/* -dictionary */
		sortInfo.sortMode = SORTMODE_DICTIONARY;
		break;
	    case 4:			/* -increasing */
		sortInfo.isIncreasing = 1;
		break;
	    case 5:			/* -index */
		if (i == (objc-2)) {
		    Tcl_StringObjAppend(resultPtr,
			    "\"-index\" option must be followed by list index", -1);
		    return TCL_ERROR;
		}
		if (Tcl_GetIntFromObj(interp, objv[i+1], &sortInfo.index)
			!= TCL_OK) {
		    return TCL_ERROR;
		}
		cmdPtr = objv[i+1];
		i++;
		break;
	    case 6:			/* -integer */
		sortInfo.sortMode = SORTMODE_INTEGER;
		break;
	    case 7:			/* -real */
		sortInfo.sortMode = SORTMODE_REAL;
		break;
	}
    }
    if (sortInfo.sortMode == SORTMODE_COMMAND) {
	Tcl_DStringInit(&sortInfo.compareCmd);
	Tcl_DStringAppend(&sortInfo.compareCmd,
		Tcl_GetStringFromObj(cmdPtr, &dummy), -1);
    }

    sortInfo.resultCode = Tcl_ListObjGetElements(interp, objv[objc-1],
	    &length, &listObjPtrs);
    if (sortInfo.resultCode != TCL_OK) {
	goto done;
    }
    if (length <= 0) {
        return TCL_OK;
    }
    elementArray = (SortElement *) ckalloc(length * sizeof(SortElement));
    for (i=0; i < length; i++){
	elementArray[i].objPtr = listObjPtrs[i];
	elementArray[i].nextPtr = &elementArray[i+1];
    }
    elementArray[length-1].nextPtr = NULL;
    elementPtr = MergeSort(elementArray, &sortInfo);
    if (sortInfo.resultCode == TCL_OK) {
	/*
	 * Note: must clear the interpreter's result object: it could
	 * have been set by the -command script.
	 */

	Tcl_ResetObjResult(interp);
	resultPtr = Tcl_GetObjResult(interp);
	for (; elementPtr != NULL; elementPtr = elementPtr->nextPtr){
	    Tcl_ListObjAppendElement(interp, resultPtr, elementPtr->objPtr);
	}
    }
    ckfree((char*) elementArray);

    done:
    if (sortInfo.sortMode == SORTMODE_COMMAND) {
	Tcl_DStringFree(&sortInfo.compareCmd);
    }
    return sortInfo.resultCode;
}

/*
 *----------------------------------------------------------------------
 *
 * MergeSort -
 *
 *	This procedure sorts a linked list of SortElement structures
 *	use the merge-sort algorithm.
 *
 * Results:
 *      A pointer to the head of the list after sorting is returned.
 *
 * Side effects:
 *	None, unless a user-defined comparison command does something
 *	weird.
 *
 *----------------------------------------------------------------------
 */

static SortElement *
MergeSort(headPtr, infoPtr)
    SortElement *headPtr;               /* First element on the list */
    SortInfo *infoPtr;                  /* Information needed by the
                                         * comparison operator */
{
    /*
     * The subList array below holds pointers to temporary lists built
     * during the merge sort.  Element i of the array holds a list of
     * length 2**i.
     */

#   define NUM_LISTS 30
    SortElement *subList[NUM_LISTS];
    SortElement *elementPtr;
    int i;

    for(i = 0; i < NUM_LISTS; i++){
        subList[i] = NULL;
    }
    while (headPtr != NULL) {
	elementPtr = headPtr;
	headPtr = headPtr->nextPtr;
	elementPtr->nextPtr = 0;
	for (i = 0; (i < NUM_LISTS) && (subList[i] != NULL); i++){
	    elementPtr = MergeLists(subList[i], elementPtr, infoPtr);
	    subList[i] = NULL;
	}
	if (i >= NUM_LISTS) {
	    i = NUM_LISTS-1;
	}
	subList[i] = elementPtr;
    }
    elementPtr = NULL;
    for (i = 0; i < NUM_LISTS; i++){
        elementPtr = MergeLists(subList[i], elementPtr, infoPtr);
    }
    return elementPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * MergeLists -
 *
 *	This procedure combines two sorted lists of SortElement structures
 *	into a single sorted list.
 *
 * Results:
 *      The unified list of SortElement structures.
 *
 * Side effects:
 *	None, unless a user-defined comparison command does something
 *	weird.
 *
 *----------------------------------------------------------------------
 */

static SortElement *
MergeLists(leftPtr, rightPtr, infoPtr)
    SortElement *leftPtr;               /* First list to be merged; may be
					 * NULL. */
    SortElement *rightPtr;              /* Second list to be merged; may be
					 * NULL. */
    SortInfo *infoPtr;                  /* Information needed by the
                                         * comparison operator. */
{
    SortElement *headPtr;
    SortElement *tailPtr;

    if (leftPtr == NULL) {
        return rightPtr;
    }
    if (rightPtr == NULL) {
        return leftPtr;
    }
    if (SortCompare(leftPtr->objPtr, rightPtr->objPtr, infoPtr) > 0) {
	tailPtr = rightPtr;
	rightPtr = rightPtr->nextPtr;
    } else {
	tailPtr = leftPtr;
	leftPtr = leftPtr->nextPtr;
    }
    headPtr = tailPtr;
    while ((leftPtr != NULL) && (rightPtr != NULL)) {
	if (SortCompare(leftPtr->objPtr, rightPtr->objPtr, infoPtr) > 0) {
	    tailPtr->nextPtr = rightPtr;
	    tailPtr = rightPtr;
	    rightPtr = rightPtr->nextPtr;
	} else {
	    tailPtr->nextPtr = leftPtr;
	    tailPtr = leftPtr;
	    leftPtr = leftPtr->nextPtr;
	}
    }
    if (leftPtr != NULL) {
       tailPtr->nextPtr = leftPtr;
    } else {
       tailPtr->nextPtr = rightPtr;
    }
    return headPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * SortCompare --
 *
 *	This procedure is invoked by MergeLists to determine the proper
 *	ordering between two elements.
 *
 * Results:
 *      A negative results means the the first element comes before the
 *      second, and a positive results means that the second element
 *      should come first.  A result of zero means the two elements
 *      are equal and it doesn't matter which comes first.
 *
 * Side effects:
 *	None, unless a user-defined comparison command does something
 *	weird.
 *
 *----------------------------------------------------------------------
 */

static int
SortCompare(objPtr1, objPtr2, infoPtr)
    Tcl_Obj *objPtr1, *objPtr2;		/* Values to be compared. */
    SortInfo *infoPtr;                  /* Information passed from the
                                         * top-level "lsort" command */
{
    int order, dummy;
    Tcl_Obj *objPtr, *resultPtr;
    char buffer[30];

    order = 0;
    if (infoPtr->resultCode != TCL_OK) {
	/*
	 * Once an error has occurred, skip any future comparisons
	 * so as to preserve the error message in sortInterp->result.
	 */

	return order;
    }
    if (infoPtr->index >= 0) {
	/*
	 * The "-index" option was specified.  Treat each object as a
	 * list, extract the requested element from each list, and
	 * compare the elements, not the lists.
	 */

	if (Tcl_ListObjIndex(infoPtr->interp, objPtr1, infoPtr->index, &objPtr)
		!= TCL_OK) {
	    infoPtr->resultCode = TCL_ERROR;
	    return order;
	}
	if (objPtr == NULL) {
	    objPtr = objPtr1;
	    missingElement:
	    resultPtr = Tcl_GetObjResult(infoPtr->interp);
	    sprintf(buffer, "%ld", infoPtr->index);
	    Tcl_StringObjAppend(resultPtr, "element ", -1);
	    Tcl_StringObjAppend(resultPtr, buffer, -1);
	    Tcl_StringObjAppend(resultPtr, " missing from sublist \"",
		    -1);
	    Tcl_StringObjAppendObj(resultPtr, objPtr);
	    Tcl_StringObjAppend(resultPtr, "\"", -1);
	    infoPtr->resultCode = TCL_ERROR;
	    return order;
	}
	objPtr1 = objPtr;
	if (Tcl_ListObjIndex(infoPtr->interp, objPtr2, infoPtr->index, &objPtr)
		!= TCL_OK) {
	    infoPtr->resultCode = TCL_ERROR;
	    return order;
	}
	if (objPtr == NULL) {
	    objPtr = objPtr2;
	    goto missingElement;
	}
	objPtr2 = objPtr;
    }
    if (infoPtr->sortMode == SORTMODE_ASCII) {
	order = strcmp(Tcl_GetStringFromObj(objPtr1, &dummy),
		Tcl_GetStringFromObj(objPtr2, &dummy));
    } else if (infoPtr->sortMode == SORTMODE_DICTIONARY) {
	order = DictionaryCompare(
		Tcl_GetStringFromObj(objPtr1, &dummy),
		Tcl_GetStringFromObj(objPtr2, &dummy));
    } else if (infoPtr->sortMode == SORTMODE_INTEGER) {
	long a, b;

	if ((Tcl_GetIntFromObj(infoPtr->interp, objPtr1, &a) != TCL_OK)
		|| (Tcl_GetIntFromObj(infoPtr->interp, objPtr2, &b)
		!= TCL_OK)) {
	    infoPtr->resultCode = TCL_ERROR;
	    return order;
	}
	if (a > b) {
	    order = 1;
	} else if (b > a) {
	    order = -1;
	}
    } else if (infoPtr->sortMode == SORTMODE_REAL) {
	double a, b;

	if ((Tcl_GetDoubleFromObj(infoPtr->interp, objPtr1, &a) != TCL_OK)
	      || (Tcl_GetDoubleFromObj(infoPtr->interp, objPtr2, &b)
		      != TCL_OK)) {
	    infoPtr->resultCode = TCL_ERROR;
	    return order;
	}
	if (a > b) {
	    order = 1;
	} else if (b > a) {
	    order = -1;
	}
    } else {
	int oldLength;
	long l;

	/*
	 * Generate and evaluate a command to determine which string comes
	 * first.
	 */

	oldLength = Tcl_DStringLength(&infoPtr->compareCmd);
	Tcl_DStringAppendElement(&infoPtr->compareCmd,
		Tcl_GetStringFromObj(objPtr1, &dummy));
	Tcl_DStringAppendElement(&infoPtr->compareCmd,
		Tcl_GetStringFromObj(objPtr2, &dummy));
	infoPtr->resultCode = Tcl_Eval(infoPtr->interp, 
		Tcl_DStringValue(&infoPtr->compareCmd));
	Tcl_DStringTrunc(&infoPtr->compareCmd, oldLength);
	if (infoPtr->resultCode != TCL_OK) {
	    Tcl_AddErrorInfo(infoPtr->interp,
		    "\n    (-compare command)");
	    return order;
	}

	/*
	 * Parse the result of the command.
	 */

	if (Tcl_GetIntFromObj(infoPtr->interp,
		Tcl_GetObjResult(infoPtr->interp), &l) != TCL_OK) {
	    TclSetObjResultFromString(infoPtr->interp,
		    "-compare command returned non-numeric result", -1);
	    infoPtr->resultCode = TCL_ERROR;
	    return order;
	}
	order = l;
    }
    if (!infoPtr->isIncreasing) {
	order = -order;
    }
    return order;
}

/*
 *----------------------------------------------------------------------
 *
 * DictionaryCompare
 *
 *	This function compares two strings as if they were being used in
 *	an index or card catalog.  The case of alphabetic characters is
 *	ignored, except to break ties.  Thus "B" comes before "b" but
 *	after "a".  Also, integers embedded in the strings compare in
 *	numerical order.  In other words, "x10y" comes after "x9y", not
 *      before it as it would when using strcmp().
 *
 * Results:
 *      A negative results means the the first element comes before the
 *      second, and a positive results means that the second element
 *      should come first.  A result of zero means the two elements
 *      are equal and it doesn't matter which comes first.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
DictionaryCompare(left, right)
    char *left, *right;          /* The strings to compare */
{
    int diff, v1, v2;
    int secondaryDiff = 0;

    while (1) {
	if (isdigit(UCHAR(*right)) && isdigit(UCHAR(*left))) {
	    v1 = strtol(left, &left, 10);
	    v2 = strtol(right, &right, 10);
	    return v1 - v2;
	}
        diff = *left - *right;
        if (diff) {
            if (isupper(UCHAR(*left)) && islower(UCHAR(*right))) {
                diff = tolower(*left) - *right;
                if (diff) {
		    return diff;
                } else if (secondaryDiff == 0) {
		    secondaryDiff = -1;
                }
            } else if (isupper(UCHAR(*right)) && islower(UCHAR(*left))) {
                diff = *left - tolower(UCHAR(*right));
                if (diff) {
		    return diff;
                } else if (secondaryDiff == 0) {
		    secondaryDiff = 1;
                }
            } else {
                return diff;
            }
        }
        if (*left == 0) {
	    break;
	}
        left++;
        right++;
    }
    if (diff == 0) {
	diff = secondaryDiff;
    }
    return diff;
}
